"""CLI for GogoGate2 and iSmartGate devices."""
import json
import sys
from dataclasses import dataclass
from enum import Enum
from typing import Callable

import pytest
from unittest.mock import MagicMock, Mock, patch
from click.testing import CliRunner
from typing_extensions import Final

import ismartgate.cli as cli_module
from ismartgate import GogoGate2Api, ISmartGateApi
from ismartgate.cli import (
    Command,
    DeviceType,
    Option,
    cli,
    gogogate2_cli,
    ismartgate_cli,
)


class StubbedTypeEnum(Enum):
    """Test enum."""

    TYPE1 = "type1"
    TYPE2 = "type2"


@dataclass
class StubbedClassData:
    """Test data class."""

    name: str
    type: StubbedTypeEnum


@patch("ismartgate.cli.GogoGate2Api")
def test_info(class_mock: Mock) -> None:
    """Test get device info."""
    api: Final = MagicMock(spec=GogoGate2Api)
    class_mock.return_value = api
    api.async_info.return_value = StubbedClassData(
        name="my_name", type=StubbedTypeEnum.TYPE1
    )

    runner: Final = CliRunner()
    result = runner.invoke(
        cli,
        (
            Option.HOST.value,
            "my_host",
            Option.USERNAME.value,
            "my_username",
            Option.PASSWORD.value,
            "my_password",
            Option.DEVICE_TYPE.value,
            DeviceType.GOGOGATE2.value,
            Command.INFO.value,
        ),
    )
    assert result.exit_code == 0
    assert json.loads(result.output) == json.loads(
        """
        {
            "name": "my_name",
            "type": "type1"
        }
    """
    )


@patch("ismartgate.cli.GogoGate2Api")
@patch("ismartgate.cli.getpass")
def test_open_with_stdin_password(getpass: Mock, class_mock: Mock) -> None:
    """Test open door."""
    api: Final = MagicMock(spec=GogoGate2Api)
    class_mock.return_value = api
    api.async_open_door.return_value = True

    getpass.return_value = "my_password"

    runner: Final = CliRunner()
    result = runner.invoke(
        cli,
        (
            Option.HOST.value,
            "my_host",
            Option.USERNAME.value,
            "my_username",
            Option.PASSWORD.value,
            "-",
            Option.DEVICE_TYPE.value,
            DeviceType.GOGOGATE2.value,
            Command.OPEN.value,
            "1",
        ),
    )
    assert result.exit_code == 0
    assert json.loads(result.output) is True


@patch("ismartgate.cli.ISmartGateApi")
@patch("ismartgate.cli.getpass")
def test_close_without_password_option(getpass: Mock, class_mock: Mock) -> None:
    """Test close door."""
    api: Final = MagicMock(spec=ISmartGateApi)
    class_mock.return_value = api
    api.async_close_door.return_value = False

    getpass.return_value = "my_password"

    runner: Final = CliRunner()
    result = runner.invoke(
        cli,
        args=(
            Option.HOST.value,
            "my_host",
            Option.USERNAME.value,
            "my_username",
            Option.DEVICE_TYPE.value,
            DeviceType.ISMARTGATE.value,
            Command.CLOSE.value,
            "1",
        ),
    )
    if result.exception:
        raise result.exception
    assert result.exit_code == 0
    assert json.loads(result.output) is False


@pytest.mark.parametrize(("cli_func",), ((gogogate2_cli,), (ismartgate_cli,)))
def test_device_cli(cli_func: Callable) -> None:
    """Test specific CLI call doesn't crash.."""
    with patch.object(sys, "argv", ["--help"]), patch.object(sys, "exit"):
        cli_func()


@patch.object(cli_module, "__name__", "__main__")
@patch.object(cli_module, "cli")
def test_maybe_init_is_main(cli_mock: Mock) -> None:
    """Test cli called when file is main."""
    cli_module.maybe_init()
    cli_mock.assert_called()


@patch.object(cli_module, "cli")
def test_maybe_init_is_not_main(cli_mock: Mock) -> None:
    """Test cli not called when file is not main."""
    cli_module.maybe_init()
    cli_mock.assert_not_called()



"""Tets for main API."""
from datetime import datetime
from typing import Callable, Union
from unittest.mock import patch

import httpx
import pytest
import respx
from typing_extensions import Final

from ismartgate import (
    AbstractGateApi,
    GogoGate2Api,
    GogoGate2ApiCipher,
    ISmartGateApi,
    ISmartGateApiCipher,
)
from ismartgate.common import (
    ApiError,
    DoorStatus,
    InvalidDoorException,
    TransitionDoorStatus,
    get_door_by_id,
)
from ismartgate.const import GogoGate2ApiErrorCode, ISmartGateApiErrorCode

from .common import MockGogoGate2Server, MockISmartGateMiniServer, MockISmartGateServer

ApiType = Union[GogoGate2Api, ISmartGateApi]
ServerType = Union[MockGogoGate2Server, MockISmartGateServer, MockISmartGateMiniServer]
ApiGenerator = Callable[[str, str, str], ApiType]
ServerGenerator = Callable[..., ServerType]


def test_ismartgate_token() -> None:
    """Test token generation."""

    assert (
        ISmartGateApiCipher("admin", "password").token
        == "f7001ecfe4d09ea0e58cb09058ba11ffe3ea36f0"
    )
    assert (
        ISmartGateApiCipher("admin2", "password2").token
        == "6669032e03454ecd1c6ea59abe49c6ed4303ff63"
    )


def test_ismartgate_encrypt_decrypt() -> None:
    """Test encrypt/decrypt."""
    cipher: Final = ISmartGateApiCipher("admin", "password")

    # Test encrypted contents matches exactly and can be decrypted.
    enc: Final = cipher.encrypt(
        '["admin", "notRealPassword", "info", "", ""]', init_vector="497c04879e0d26af"
    )
    assert (
        enc
        == "497c04879e0d26afxuTQ0lB1Rd0c0G/l6Tiw+YCjnN9oG26d3I5IyGQpvkcpJ9l2aHDcTdquB0RnkWgi"
    )
    assert cipher.decrypt(enc) == '["admin", "notRealPassword", "info", "", ""]'

    # Test with generated initialization vector.
    assert cipher.decrypt(cipher.encrypt("Hello World")) == "Hello World"

    # Test initialization vector padding.
    assert (
        cipher.decrypt(cipher.encrypt("Hello World", init_vector="A")) == "Hello World"
    )


def test_gogogate2_cipher() -> None:
    """Test encrypt/decrypt."""
    cipher: Final = GogoGate2ApiCipher()

    # Test encrypted contents matches exactly and can be decrypted.
    enc: Final = cipher.encrypt(
        '["admin", "notRealPassword", "info", "", ""]', init_vector="497c04879e0d26af"
    )
    assert (
        enc
        == "497c04879e0d26afhnGLw5SJBnC0qUqNFHYTBgzIGsTyUF96EyZMyAHIUrLiQobociSp4fkBhf1sWSWc"
    )
    assert cipher.decrypt(enc) == '["admin", "notRealPassword", "info", "", ""]'

    # Test with generated initialization vector.
    assert cipher.decrypt(cipher.encrypt("Hello World")) == "Hello World"

    # Test initialization vector padding.
    assert (
        cipher.decrypt(cipher.encrypt("Hello World", init_vector="A")) == "Hello World"
    )


@pytest.mark.parametrize(
    ("api_generator", "server_generator", "error_code"),
    (
        (
            GogoGate2Api,
            MockGogoGate2Server,
            GogoGate2ApiErrorCode.CREDENTIALS_INCORRECT.value,
        ),
        (
            ISmartGateApi,
            MockISmartGateServer,
            ISmartGateApiErrorCode.CREDENTIALS_INCORRECT.value,
        ),
        (
            ISmartGateApi,
            MockISmartGateMiniServer,
            ISmartGateApiErrorCode.CREDENTIALS_INCORRECT.value,
        ),
    ),
)
@pytest.mark.asyncio
@respx.mock
async def test_api_invalid_credentials(
    api_generator: ApiGenerator, server_generator: ServerGenerator, error_code: int
) -> None:
    """Test invalid credentials error."""
    # Incorrect username/password.
    api: Final = api_generator("device1", "fakeuser", "fakepassword")
    server_generator(api, username="fakeuser1", password="fakepassword2")
    with pytest.raises(ApiError) as exinfo:
        await api.async_info()
    assert exinfo.value.code == error_code


@pytest.mark.parametrize(
    ("api_generator", "server_generator"),
    (
        (GogoGate2Api, MockGogoGate2Server),
        (ISmartGateApi, MockISmartGateServer),
        (ISmartGateApi, MockISmartGateMiniServer),
    ),
)
@pytest.mark.asyncio
@respx.mock
async def test_activate(
    api_generator: ApiGenerator, server_generator: ServerGenerator
) -> None:
    """Test activate."""
    api: Final = api_generator("device1", "fakeuser", "fakepassword")
    server_generator(api)

    response = await api.async_activate(1)
    assert response
    assert response.result


@pytest.mark.parametrize(
    ("api_generator", "server_generator"),
    (
        (GogoGate2Api, MockGogoGate2Server),
        (ISmartGateApi, MockISmartGateServer),
    ),
)
@pytest.mark.asyncio
@respx.mock
# pylint: disable=too-many-statements
async def test_open_and_close_door(
    api_generator: ApiGenerator, server_generator: ServerGenerator
) -> None:
    """Test open and close door."""
    api: Final = api_generator("device1", "fakeuser", "fakepassword")
    server_generator(api)

    # Initial info.
    response = await api.async_info()
    door1 = get_door_by_id(1, response)
    door2 = get_door_by_id(2, response)
    door3 = get_door_by_id(3, response)
    assert door1
    assert door2
    assert door3
    assert door1.status == DoorStatus.CLOSED
    assert door2.status == DoorStatus.OPENED
    assert door3.status == DoorStatus.UNDEFINED

    # Nothing changes because door is already closed.
    assert await api.async_close_door(1) is False
    response = await api.async_info()
    door1 = get_door_by_id(1, response)
    door2 = get_door_by_id(2, response)
    door3 = get_door_by_id(3, response)
    assert door1
    assert door2
    assert door3
    assert door1.status == DoorStatus.CLOSED
    assert door2.status == DoorStatus.OPENED
    assert door3.status == DoorStatus.UNDEFINED

    # Open a door.
    assert await api.async_open_door(1) is True
    response = await api.async_info()
    door1 = get_door_by_id(1, response)
    door2 = get_door_by_id(2, response)
    door3 = get_door_by_id(3, response)
    assert door1
    assert door2
    assert door3
    assert door1.status == DoorStatus.OPENED
    assert door2.status == DoorStatus.OPENED
    assert door3.status == DoorStatus.UNDEFINED

    # Close a door.
    assert await api.async_close_door(2) is True
    response = await api.async_info()
    door1 = get_door_by_id(1, response)
    door2 = get_door_by_id(2, response)
    door3 = get_door_by_id(3, response)
    assert door1
    assert door2
    assert door3
    assert door1.status == DoorStatus.OPENED
    assert door2.status == DoorStatus.CLOSED
    assert door3.status == DoorStatus.UNDEFINED

    # No change for already closed door.
    assert await api.async_close_door(2) is False
    response = await api.async_info()
    door1 = get_door_by_id(1, response)
    door2 = get_door_by_id(2, response)
    door3 = get_door_by_id(3, response)
    assert door1
    assert door2
    assert door3
    assert door1.status == DoorStatus.OPENED
    assert door2.status == DoorStatus.CLOSED
    assert door3.status == DoorStatus.UNDEFINED

    # No change for unknown door.
    assert await api.async_close_door(8) is False
    response = await api.async_info()
    door1 = get_door_by_id(1, response)
    door2 = get_door_by_id(2, response)
    door3 = get_door_by_id(3, response)
    assert door1
    assert door2
    assert door3
    assert door1.status == DoorStatus.OPENED
    assert door2.status == DoorStatus.CLOSED
    assert door3.status == DoorStatus.UNDEFINED

    # No change for unsupported status.
    # pylint: disable=protected-access
    assert await api._async_set_door_status(1, DoorStatus.UNDEFINED) is False
    response = await api.async_info()
    door1 = get_door_by_id(1, response)
    door2 = get_door_by_id(2, response)
    door3 = get_door_by_id(3, response)
    assert door1
    assert door2
    assert door3
    assert door1.status == DoorStatus.OPENED
    assert door2.status == DoorStatus.CLOSED
    assert door3.status == DoorStatus.UNDEFINED


@pytest.mark.parametrize(
    ("api_generator", "server_generator", "true_value"),
    (
        (GogoGate2Api, MockGogoGate2Server, "1"),
        (ISmartGateApi, MockISmartGateServer, "yes"),
    ),
)
@pytest.mark.asyncio
@respx.mock
# pylint: disable=too-many-statements
async def test_remoteaccess(
    api_generator: ApiGenerator, server_generator: ServerGenerator, true_value: str
) -> None:
    """Test open and close door."""
    api: Final = api_generator("device1", "fakeuser", "fakepassword")
    server: Final = server_generator(api)

    server.set_info_value("remoteaccessenabled", "false")
    assert not (await api.async_info()).remoteaccessenabled
    server.set_info_value("remoteaccessenabled", "no")
    assert not (await api.async_info()).remoteaccessenabled
    server.set_info_value("remoteaccessenabled", "0")
    assert not (await api.async_info()).remoteaccessenabled

    server.set_info_value("remoteaccessenabled", true_value)
    assert (await api.async_info()).remoteaccessenabled
    server.set_info_value("remoteaccessenabled", true_value.upper())
    assert (await api.async_info()).remoteaccessenabled
    server.set_info_value("remoteaccessenabled", true_value.lower())
    assert (await api.async_info()).remoteaccessenabled


@pytest.mark.parametrize(
    ("api_generator", "server_generator"),
    (
        (GogoGate2Api, MockGogoGate2Server),
        (ISmartGateApi, MockISmartGateServer),
    ),
)
@pytest.mark.asyncio
@respx.mock
# pylint: disable=too-many-statements
async def test_sensor_temperature_and_voltage(
    api_generator: ApiGenerator, server_generator: ServerGenerator
) -> None:
    """Test open and close door."""
    api: Final = api_generator("device1", "fakeuser", "fakepassword")
    server_generator(api)

    # Initial info.
    response: Final = await api.async_info()
    assert response.door1.temperature == 16.3
    assert response.door1.voltage == 40

    assert response.door2.temperature is None
    assert response.door2.voltage == 40

    assert response.door3.temperature == 16.3
    assert response.door3.voltage is None


@pytest.mark.asyncio
@respx.mock
# pylint: disable=too-many-statements
async def test_empty_apicode() -> None:
    """Test open and close door."""
    api: Final = ISmartGateApi("device1", "fakeuser", "fakepassword")
    MockISmartGateServer(api, "")

    # Initial info.
    response: Final = await api.async_info()
    assert response.door1.apicode == ""
    assert response.door2.apicode == "0"
    assert response.door3.apicode == "0"


@pytest.mark.parametrize(
    ("api_generator", "server_generator"),
    (
        (GogoGate2Api, MockGogoGate2Server),
        (ISmartGateApi, MockISmartGateServer),
    ),
)
@pytest.mark.asyncio
@respx.mock
# pylint: disable=too-many-statements
async def test_transitional_door_statuses(
    api_generator: ApiGenerator, server_generator: ServerGenerator
) -> None:
    """Test open and close door."""
    with patch("ismartgate.datetime") as datetime_mock:
        api: Final = api_generator("device1", "fakeuser", "fakepassword")
        mock_server: Final = server_generator(api)

        # Test current status.
        datetime_mock.utcnow.side_effect = datetime.utcnow
        api._transition_door_status.clear()  # pylint: disable=protected-access
        info = await api.async_info()
        assert await api.async_get_door_statuses() == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert await api.async_get_door_statuses(use_transitional_status=True) == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert await api.async_get_door_statuses(use_transitional_status=False) == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert api.async_get_door_statuses_from_info(info) == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert api.async_get_door_statuses_from_info(
            info, use_transitional_status=True
        ) == {1: DoorStatus.CLOSED, 2: DoorStatus.OPENED}
        assert api.async_get_door_statuses_from_info(
            info, use_transitional_status=False
        ) == {1: DoorStatus.CLOSED, 2: DoorStatus.OPENED}

        # Test door is in the process of opening.
        datetime_mock.utcnow.side_effect = datetime.utcnow
        api._transition_door_status.clear()  # pylint: disable=protected-access
        await api.async_open_door(1)
        mock_server.set_device_status(1, DoorStatus.CLOSED)
        info = await api.async_info()
        assert await api.async_get_door_statuses() == {
            1: TransitionDoorStatus.OPENING,
            2: DoorStatus.OPENED,
        }
        assert await api.async_get_door_statuses(use_transitional_status=True) == {
            1: TransitionDoorStatus.OPENING,
            2: DoorStatus.OPENED,
        }
        assert await api.async_get_door_statuses(use_transitional_status=False) == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert api.async_get_door_statuses_from_info(info) == {
            1: TransitionDoorStatus.OPENING,
            2: DoorStatus.OPENED,
        }
        assert api.async_get_door_statuses_from_info(
            info, use_transitional_status=True
        ) == {1: TransitionDoorStatus.OPENING, 2: DoorStatus.OPENED}
        assert api.async_get_door_statuses_from_info(
            info, use_transitional_status=False
        ) == {1: DoorStatus.CLOSED, 2: DoorStatus.OPENED}

        # Door is open before the transitional cache timeout.
        datetime_mock.utcnow.side_effect = datetime.utcnow
        api._transition_door_status.clear()  # pylint: disable=protected-access
        await api.async_open_door(1)
        info = await api.async_info()
        assert await api.async_get_door_statuses() == {
            1: DoorStatus.OPENED,
            2: DoorStatus.OPENED,
        }
        assert await api.async_get_door_statuses(use_transitional_status=True) == {
            1: DoorStatus.OPENED,
            2: DoorStatus.OPENED,
        }
        assert await api.async_get_door_statuses(use_transitional_status=False) == {
            1: DoorStatus.OPENED,
            2: DoorStatus.OPENED,
        }
        assert api.async_get_door_statuses_from_info(info) == {
            1: DoorStatus.OPENED,
            2: DoorStatus.OPENED,
        }
        assert api.async_get_door_statuses_from_info(
            info, use_transitional_status=True
        ) == {1: DoorStatus.OPENED, 2: DoorStatus.OPENED}
        assert api.async_get_door_statuses_from_info(
            info, use_transitional_status=False
        ) == {1: DoorStatus.OPENED, 2: DoorStatus.OPENED}

        # Door remains closed after the transitional cache timeout.
        datetime_mock.utcnow.side_effect = datetime.utcnow
        api._transition_door_status.clear()  # pylint: disable=protected-access
        mock_server.set_device_status(1, DoorStatus.CLOSED)
        await api.async_open_door(1)
        mock_server.set_device_status(1, DoorStatus.CLOSED)
        datetime_mock.utcnow.side_effect = (
            lambda: datetime.utcnow()
            + AbstractGateApi.DEFAULT_TRANSITION_STATUS_TIMEOUT
        )
        info = await api.async_info()
        assert await api.async_get_door_statuses() == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert await api.async_get_door_statuses(use_transitional_status=True) == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert await api.async_get_door_statuses(use_transitional_status=False) == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert api.async_get_door_statuses_from_info(info) == {
            1: DoorStatus.CLOSED,
            2: DoorStatus.OPENED,
        }
        assert api.async_get_door_statuses_from_info(
            info, use_transitional_status=True
        ) == {1: DoorStatus.CLOSED, 2: DoorStatus.OPENED}
        assert api.async_get_door_statuses_from_info(
            info, use_transitional_status=False
        ) == {1: DoorStatus.CLOSED, 2: DoorStatus.OPENED}


@pytest.mark.parametrize(
    ("api_generator", "server_generator"),
    ((ISmartGateApi, MockISmartGateServer), (ISmartGateApi, MockISmartGateMiniServer)),
)
@pytest.mark.asyncio
@respx.mock
# pylint: disable=too-many-statements
async def test_ismartgate_activate_invalid_door(
    api_generator: ApiGenerator, server_generator: ServerGenerator
) -> None:
    """Test invalid door exception when calling activate for ismartgate."""
    api: Final = api_generator("device1", "fakeuser", "fakepassword")
    server_generator(api)

    with pytest.raises(InvalidDoorException):
        await api.async_activate(5)


@pytest.mark.parametrize(
    ("api_generator", "server_generator"),
    (
        (GogoGate2Api, MockGogoGate2Server),
        (ISmartGateApi, MockISmartGateServer),
        (ISmartGateApi, MockISmartGateMiniServer),
    ),
)
@pytest.mark.asyncio
@respx.mock
# pylint: disable=too-many-statements
async def test_httpx_client(
    api_generator: ApiGenerator, server_generator: ServerGenerator
) -> None:
    """Verify we can accept a shared httpx AsyncClient."""
    httpx_async_client: httpx.AsyncClient = httpx.AsyncClient()
    api: Final = api_generator(  # type: ignore
        "device1", "fakeuser", "fakepassword", httpx_async_client=httpx_async_client
    )
    server_generator(api)

    assert await api.async_info()



"""Tests for common code."""
import pytest
from defusedxml import ElementTree
from typing_extensions import Final

from ismartgate.common import (
    DoorMode,
    DoorStatus,
    GogoGate2Door,
    GogoGate2InfoResponse,
    Network,
    Outputs,
    TagNotFoundException,
    TextEmptyException,
    UnexpectedTypeException,
    Wifi,
    element_text_or_raise,
    enum_or_raise,
    get_configured_door_by_id,
    get_configured_doors,
    int_or_raise,
    str_or_raise,
)


def test_element_exceptions() -> None:
    """Test exceptions thrown while parsing xml elements."""
    root_element: Final = ElementTree.fromstring(
        """
            <response>
                <tag1></tag1>
                <tag2>value</tag2>
            </response>
        """
    )

    with pytest.raises(TagNotFoundException) as exinfo1:
        element_text_or_raise(root_element, "tag3")
        assert exinfo1.value.tag == "tag3"

    with pytest.raises(TextEmptyException) as exinfo2:
        element_text_or_raise(root_element, "tag1")
        assert exinfo2.value.tag == "tag1"

    with pytest.raises(UnexpectedTypeException) as exinfo3:
        int_or_raise(element_text_or_raise(root_element, "tag2"))
        assert exinfo3.value.value == "value"
        assert exinfo3.value.expected == int


def test_str_or_raise() -> None:
    """Test exceptions for strings."""
    with pytest.raises(UnexpectedTypeException) as exinfo:
        str_or_raise(None)
        assert exinfo.value.value is None
        assert exinfo.value.expected == str

    assert str_or_raise(123) == "123"


def test_enum_or_raise() -> None:
    """Test exceptions for enums."""
    with pytest.raises(UnexpectedTypeException) as exinfo:
        enum_or_raise(None, DoorStatus)
        assert exinfo.value.value is None
        assert exinfo.value.expected == DoorStatus


def test_get_configured_doors() -> None:
    """Test get configurd doors."""
    response: Final = GogoGate2InfoResponse(
        user="user1",
        gogogatename="gogogatename1",
        model="",
        apiversion="",
        remoteaccessenabled=False,
        remoteaccess="",
        firmwareversion="",
        apicode="",
        door1=GogoGate2Door(
            door_id=1,
            permission=True,
            name="Door1",
            gate=False,
            mode=DoorMode.GARAGE,
            status=DoorStatus.OPENED,
            sensor=True,
            sensorid=None,
            camera=False,
            events=2,
            temperature=None,
            voltage=None,
        ),
        door2=GogoGate2Door(
            door_id=2,
            permission=True,
            name=None,
            gate=True,
            mode=DoorMode.GARAGE,
            status=DoorStatus.OPENED,
            sensor=True,
            sensorid=None,
            camera=False,
            events=2,
            temperature=None,
            voltage=None,
        ),
        door3=GogoGate2Door(
            door_id=3,
            permission=True,
            name="Door3",
            gate=False,
            mode=DoorMode.GARAGE,
            status=DoorStatus.OPENED,
            sensor=True,
            sensorid=None,
            camera=False,
            events=2,
            temperature=None,
            voltage=None,
        ),
        outputs=Outputs(output1=True, output2=False, output3=True),
        network=Network(ip=""),
        wifi=Wifi(SSID="", linkquality="", signal=""),
    )

    assert get_configured_door_by_id(1, response) == response.door1
    assert not get_configured_door_by_id(2, response)
    assert get_configured_door_by_id(3, response) == response.door3
    assert get_configured_doors(response) == (response.door1, response.door3)



"""Common test code."""
import abc
import json
from typing import Any, Generic, Optional, TypeVar, Union
from urllib.parse import parse_qs
from xml.dom.minidom import parseString

import dicttoxml
import respx
from httpx import Request, Response
from typing_extensions import Final

from ismartgate import AbstractGateApi, ISmartGateApiCipher
from ismartgate.common import (
    DoorMode,
    DoorStatus,
    EnhancedJSONEncoder,
    GogoGate2Door,
    GogoGate2InfoResponse,
    ISmartGateDoor,
    ISmartGateInfoResponse,
    Network,
    Outputs,
    RequestOption,
    Wifi,
)
from ismartgate.const import NONE_INT, GogoGate2ApiErrorCode, ISmartGateApiErrorCode

MockInfoResponse = TypeVar(
    "MockInfoResponse", bound=Union[GogoGate2InfoResponse, ISmartGateInfoResponse]
)


class AbstractMockServer(Generic[MockInfoResponse], abc.ABC):
    """Mock server."""

    def __init__(
        self,
        api: AbstractGateApi,
        api_code: str = "api_code1",
        host: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ):
        """Init object."""
        self.host: Final = host or api.host
        self.username: Final = username or api.username
        self.password: Final = password or api.password
        self.api_code: Final = api_code
        self.http_status: Final = 200
        self.cipher: Final = api.cipher
        self._info_data: dict = json.loads(
            json.dumps(self._get_info_data(), indent=2, cls=EnhancedJSONEncoder)
        )

        respx.reset()
        respx.get(AbstractGateApi.API_URL_TEMPLATE % self.host).mock(
            side_effect=self._handle_request
        )

    @abc.abstractmethod
    def _get_info_data(self) -> MockInfoResponse:
        pass

    @abc.abstractmethod
    def _get_error_corrupted_data_response(self) -> Response:
        pass

    @abc.abstractmethod
    def _get_error_token_not_set_response(self) -> Response:
        pass

    @abc.abstractmethod
    def _get_error_invalid_token_response(self) -> Response:
        pass

    @abc.abstractmethod
    def _get_error_absent_credentials_response(self) -> Response:
        pass

    @abc.abstractmethod
    def _get_error_invalid_credentials_response(self) -> Response:
        pass

    @abc.abstractmethod
    def _get_error_invalid_option_response(self) -> Response:
        pass

    @abc.abstractmethod
    def _get_error_activate_invalid_api_code_response(self) -> Response:
        pass

    @abc.abstractmethod
    def _get_error_activate_door_id_not_set_response(self) -> Response:
        pass

    @abc.abstractmethod
    def _get_error_activate_invalid_door_response(self) -> Response:
        pass

    def set_device_status(self, door_id: int, door_status: DoorStatus) -> None:
        """Set the status of a device."""
        self._info_data[f"door{door_id}"]["status"] = door_status.value

    def set_info_value(self, name: str, value: Any) -> None:
        """Set a value of info data."""
        self._info_data[name] = value

    # pylint: disable=too-many-return-statements
    def _handle_request(self, request: Request) -> Response:
        # Simulate an HTTP error.
        if self.http_status != 200:
            return self._new_response("")

        # Parse the request.
        query: Final = parse_qs(request.url.query)
        data: Final = query[b"data"][0].decode()

        try:
            decrypted: Final = self.cipher.decrypt(data)
            payload: Final = json.loads(decrypted)
            username: Final = payload[0]
            password: Final = payload[1]
            option: Final = payload[2]
            arg1: Final = payload[3]
            arg2: Final = payload[4]
        except Exception as ex:  # pylint: disable=broad-except
            print(ex)
            return self._get_error_corrupted_data_response()

        # Maybe validate token.
        if isinstance(self.cipher, ISmartGateApiCipher):
            if b"token" not in query:
                return self._get_error_token_not_set_response()

            # API returns invalid credentials when token is wrong
            token: Final = query[b"token"][0].decode()
            if token != self.cipher.token:
                return self._get_error_invalid_token_response()

        # Validate credentials.
        if not username or not password:
            return self._get_error_absent_credentials_response()

        if username != self.username or password != self.password:
            return self._get_error_invalid_credentials_response()

        if option not in [item.value for item in RequestOption]:
            return self._get_error_invalid_option_response()

        if option == RequestOption.INFO.value:
            return self._info_response()

        if option == RequestOption.ACTIVATE.value:
            return self._activate_response(arg1, arg2)

        return self._okay_response()

    def _okay_response(self) -> Response:
        return self._new_response({"result": "OK"})

    def _activate_response(self, door_id: str, api_code: str) -> Response:
        if api_code != self.api_code:
            return self._get_error_activate_invalid_api_code_response()

        if not door_id.isdigit():
            return self._get_error_activate_door_id_not_set_response()

        door: Final = self._info_data[f"door{door_id}"]
        if not door:
            return self._get_error_activate_invalid_door_response()

        # This door is not setup.
        if not door["name"]:
            return self._okay_response()

        door["status"] = (
            DoorStatus.CLOSED.value
            if door["status"] == DoorStatus.OPENED.value
            else DoorStatus.OPENED.value
        )

        return self._okay_response()

    def _info_response(self) -> Response:
        return self._new_response(self._info_data)

    def _error_response(self, code: int, message: str) -> Response:
        return self._new_response(
            {"error": {"errorcode": code, "errormsg": f"Code: {code} - {message}"}},
            encrypt=False,
        )

    def _new_response(self, data: Any, encrypt: bool = True) -> Response:
        xml_str = parseString(
            dicttoxml.dicttoxml(data, custom_root="response", attr_type=False)
        ).toprettyxml()

        return Response(
            self.http_status,
            content=self.cipher.encrypt(xml_str) if encrypt else xml_str,
        )


class MockGogoGate2Server(AbstractMockServer[GogoGate2InfoResponse]):
    """Test server for GogoGate2"""

    def _get_info_data(self) -> GogoGate2InfoResponse:
        return GogoGate2InfoResponse(
            user=self.username,
            model="GG2",
            apicode=self.api_code,
            apiversion="apiversion123",
            remoteaccessenabled=False,
            remoteaccess="abcdefg12345.my-gogogate.com",
            firmwareversion="761",
            gogogatename="Home",
            door1=GogoGate2Door(
                door_id=1,
                permission=True,
                name="My Door 1",
                gate=False,
                status=DoorStatus.CLOSED,
                mode=DoorMode.GARAGE,
                sensor=True,
                camera=False,
                events=None,
                sensorid="sensor123",
                temperature=16.3,
                voltage=40,
            ),
            door2=GogoGate2Door(
                door_id=3,
                permission=True,
                name="My Door 2",
                gate=True,
                status=DoorStatus.OPENED,
                mode=DoorMode.GARAGE,
                sensor=False,
                camera=False,
                events=None,
                sensorid="sensor123",
                temperature=NONE_INT,
                voltage=40,
            ),
            door3=GogoGate2Door(
                door_id=3,
                permission=True,
                name=None,
                gate=False,
                status=DoorStatus.UNDEFINED,
                mode=DoorMode.GARAGE,
                sensor=False,
                camera=False,
                events=None,
                sensorid="sensor123",
                temperature=16.3,
                voltage=NONE_INT,
            ),
            outputs=Outputs(
                output1=True,
                output2=False,
                output3=False,
            ),
            network=Network(ip="127.0.0.1"),
            wifi=Wifi(SSID="Wifi network", linkquality="80%", signal="20"),
        )

    def _get_error_corrupted_data_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.CORRUPTED_DATA.value, "Error: corrupted data"
        )

    def _get_error_invalid_token_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.INVALID_TOKEN.value,
            "Error: invalid token",
        )

    def _get_error_token_not_set_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.TOKEN_NOT_SET.value,
            "Error: token not set",
        )

    def _get_error_absent_credentials_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.CREDENTIALS_NOT_SET.value,
            "Error: login or password not set",
        )

    def _get_error_invalid_credentials_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.CREDENTIALS_INCORRECT.value,
            "Error: wrong login or password",
        )

    def _get_error_invalid_option_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.INVALID_OPTION.value, "Error: invalid option"
        )

    def _get_error_activate_invalid_api_code_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.INVALID_API_CODE.value, "Error: invalid API code"
        )

    def _get_error_activate_door_id_not_set_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.DOOR_NOT_SET.value, "Error: door not set"
        )

    def _get_error_activate_invalid_door_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.INVALID_DOOR.value, "Error: invalid door"
        )


class MockISmartGateServer(AbstractMockServer[ISmartGateInfoResponse]):
    """Test server for ISmartGate"""

    def _get_info_data(self) -> ISmartGateInfoResponse:
        return ISmartGateInfoResponse(
            pin=1234,
            lang="en",
            ismartgatename="Home",
            newfirmware=False,
            user=self.username,
            model="GG2",
            apiversion="apiversion123",
            remoteaccessenabled=False,
            remoteaccess="abcdefg12345.my-gogogate.com",
            firmwareversion="761",
            door1=ISmartGateDoor(
                enabled=True,
                apicode=self.api_code,
                customimage=False,
                door_id=1,
                permission=True,
                name="My Door 1",
                gate=False,
                status=DoorStatus.CLOSED,
                mode=DoorMode.GARAGE,
                sensor=True,
                camera=False,
                events=None,
                sensorid="sensor123",
                temperature=16.3,
                voltage=40,
            ),
            door2=ISmartGateDoor(
                enabled=True,
                apicode=self.api_code,
                customimage=False,
                door_id=3,
                permission=True,
                name="My Door 2",
                gate=True,
                status=DoorStatus.OPENED,
                mode=DoorMode.GARAGE,
                sensor=False,
                camera=False,
                events=None,
                sensorid="sensor123",
                temperature=NONE_INT,
                voltage=40,
            ),
            door3=ISmartGateDoor(
                enabled=True,
                apicode=self.api_code,
                customimage=False,
                door_id=3,
                permission=True,
                name=None,
                gate=False,
                status=DoorStatus.UNDEFINED,
                mode=DoorMode.GARAGE,
                sensor=False,
                camera=False,
                events=None,
                sensorid="sensor123",
                temperature=16.3,
                voltage=NONE_INT,
            ),
            network=Network(ip="127.0.0.1"),
            wifi=Wifi(SSID="Wifi network", linkquality="80%", signal="20"),
        )

    def _get_error_corrupted_data_response(self) -> Response:
        return self._error_response(
            GogoGate2ApiErrorCode.CORRUPTED_DATA.value, "Error: corrupted data"
        )

    def _get_error_invalid_token_response(self) -> Response:
        return self._error_response(
            ISmartGateApiErrorCode.INVALID_TOKEN.value,
            "Error: invalid token",
        )

    def _get_error_token_not_set_response(self) -> Response:
        return self._error_response(
            ISmartGateApiErrorCode.TOKEN_NOT_SET.value,
            "Error: token not set",
        )

    def _get_error_absent_credentials_response(self) -> Response:
        return self._error_response(
            ISmartGateApiErrorCode.CREDENTIALS_NOT_SET.value,
            "Error: login or password not set",
        )

    def _get_error_invalid_credentials_response(self) -> Response:
        return self._error_response(
            ISmartGateApiErrorCode.CREDENTIALS_INCORRECT.value,
            "Error: wrong login or password",
        )

    def _get_error_invalid_option_response(self) -> Response:
        return self._error_response(
            ISmartGateApiErrorCode.INVALID_OPTION.value, "Error: invalid option"
        )

    def _get_error_activate_invalid_api_code_response(self) -> Response:
        return self._error_response(
            ISmartGateApiErrorCode.INVALID_API_CODE.value, "Error: invalid API code"
        )

    def _get_error_activate_door_id_not_set_response(self) -> Response:
        return self._error_response(
            ISmartGateApiErrorCode.DOOR_NOT_SET.value, "Error: door not set"
        )

    def _get_error_activate_invalid_door_response(self) -> Response:
        return self._error_response(
            ISmartGateApiErrorCode.INVALID_DOOR.value, "Error: invalid door"
        )


class MockISmartGateMiniServer(MockISmartGateServer):
    """Test server for ISmartGate Mini"""

    def _get_info_data(self) -> ISmartGateInfoResponse:
        return ISmartGateInfoResponse(
            pin=1234,
            lang="en",
            ismartgatename="Home",
            newfirmware=False,
            user=self.username,
            model="iSmartGate Mini",
            apiversion="apiversion123",
            remoteaccessenabled=False,
            remoteaccess="abcdefg12345.my-gogogate.com",
            firmwareversion="761",
            door1=ISmartGateDoor(
                enabled=True,
                apicode=self.api_code,
                customimage=False,
                door_id=1,
                permission=True,
                name="My Door 1",
                gate=False,
                status=DoorStatus.CLOSED,
                mode=DoorMode.GARAGE,
                sensor=True,
                camera=False,
                events=None,
                sensorid="sensor123",
                temperature=16.3,
                voltage=40,
            ),
            door2=ISmartGateDoor(
                door_id=2,
                enabled=False,
                apicode=None,
                customimage=False,
                permission=True,
                name=None,
                gate=False,
                mode=DoorMode.GARAGE,
                status=DoorStatus.UNDEFINED,
                sensorid="0",
                camera=False,
                temperature=None,
                voltage=None,
                sensor=False,
                events=None,
            ),
            door3=ISmartGateDoor(
                door_id=3,
                enabled=False,
                apicode=None,
                customimage=False,
                permission=True,
                name=None,
                gate=False,
                mode=DoorMode.GARAGE,
                status=DoorStatus.UNDEFINED,
                sensorid="0",
                camera=False,
                temperature=None,
                voltage=None,
                sensor=False,
                events=None,
            ),
            network=Network(ip="127.0.0.1"),
            wifi=Wifi(SSID="Wifi network", linkquality="80%", signal="20"),
        )
